#!/usr/bin/python3
# -*- coding: utf-8 -*-
app_svnid = '$HeadURL: https://svn.pasta.freemyip.com/main/ade/tags/3.0.1/share/templates/lxpython/bin/lxpython.py $ $LastChangedRevision: 10649 $'

#  Modules
import sys
import subprocess
sys.path.append(subprocess.Popen(["ade-config", "ade_share_prefix"], stdout=subprocess.PIPE, universal_newlines=True).communicate()[0][:-1] + '/include')
import ade
import os

#  Errors
lxpython_defined_errors  = {
    "lxpython_err_misc":{"fmt":"%s"},
}

#  Options
opt_force = None;
opt_labelwhat = None;

#  Other globals

def lxpython(errstack):
    global opt_force, opt_labelwhat

    #  Defaults for options
    opt_force = 0
    opt_labelwhat = 0

    #  Register errors, options and callbacks
    ade.register_error_types(lxpython_defined_errors)
    rc = ade.register_options(errstack, "wibF", "window,icon,both,force", globals(), "handle_option_%s")
    if rc != ade.ok:
        return rc
    rc = ade.set_callbacks(errstack, lxpython_usage_help, lxpython_version, lxpython_paths)
    if rc != ade.ok:
        return rc

    #  Process options
    rc = ade.process_options(errstack)
    if rc != ade.ok:
        return rc


    #  Process arguments
    len(sys.argv) != 1 and ade.show_bad_usage(errstack)
    label = sys.argv[0]

    #  Sanity checks and derivations
    if not opt_force and not is_relabelable_term(errstack, os.getenv('TERM'))[1]:
        ade.error(errstack, "lxpython_err_misc", "the title of this sort of terminal cannot be changed by this program (is 'TERM' set correctly?)")
        return ade.fail

    #  Guts
    sys.stdout.write("\033]%d;%s\007" % (opt_labelwhat, label))
    return ade.ok

#  Option handlers
def handle_option_w(errstack):
    global opt_labelwhat
    opt_labelwhat = 2
    return ade.ok

def handle_option_window(*args):
    return handle_option_w(*args)

def handle_option_i(errstack):
    global opt_labelwhat
    opt_labelwhat = 1
    return ade.ok

def handle_option_icon(*args):
    return handle_option_i(*args)

def handle_option_b(errstack):
    global opt_labelwhat
    opt_labelwhat = 0
    return ade.ok

def handle_option_both(*args):
    return handle_option_b(*args)

def handle_option_F(errstack):
    global opt_force
    opt_force = 1
    return ade.ok

def handle_option_force(*args):
    return handle_option_F(*args)

#  Callbacks
def lxpython_usage_help(errstack):
    return ade.ok, "<text>", "         -w        | --window                label the window only\n" + \
                                     "         -i        | --icon                  label the icon only\n" + \
                                     "         -b        | --both                  label both (default)\n" + \
                                     "         -F        | --force                 ignore 'TERM' and force labelling\n"

def lxpython_version(errstack):
    return ade.extract_version(errstack, app_svnid)

def lxpython_paths(errstack):
    return ade.ok, None

#  Other functions
def is_relabelable_term(errstack, term):
    if term is not None and term in "rxvt xterm xterm-debian cygwin".split():
        isrelabelable_flag = 1
    elif term is not None and "xterm" in term:
        ade.warning(errstack, "lxpython_err_misc", "guessing this is some sort of xterm")
        isrelabelable_flag = 1
    elif term == "vt100" and "XSC|" in os.getenv('TERMCAP'):
        isrelabelable_flag = 1
    else:
        isrelabelable_flag = 0

    #  pass up and return
    return ade.ok, isrelabelable_flag

#  Entry point
ade.main(lxpython)
