#!/bin/bash
#  $HeadURL: https://svn.pasta.freemyip.com/main/ade/tags/3.0.1/tests/bin/ADE516_stdlibs_writers.sh $ $LastChangedRevision: 10813 $

echo "setting up environment ..."
ADELANGS=( shell perl python )
CHECKS=( default devnull logfile syslog )
echo "defining support functions ..."
default_shell()
{
    adeshf '
    ade_error "$ERRSTACK_REF" ADE_ERR_MISC "this is a test"
    return $ADE_FAIL
'
}

devnull_shell()
{
    adeshf '
    ade_set_messaging_parameters "$ERRSTACK_REF" writers=devnull
    ade_error "$ERRSTACK_REF" ADE_ERR_MISC "this is a test"
    return $ADE_FAIL
'
}

logfile_shell()
{
    LOGFILE=$(mktemp)
    adeshf "
    ade_set_messaging_parameters \"\$ERRSTACK_REF\" writers=logfile logfile=$LOGFILE
    ade_error \"\$ERRSTACK_REF\" ADE_ERR_MISC \"this is a test\"
    return \$ADE_FAIL
"
    #  Show file but trim timestamps.
    sed -r 's/^.{21}//' $LOGFILE
    rm $LOGFILE
}

syslog_shell()
{
    RANDOM_TEXT=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 20)
    adeshf "
    ade_set_messaging_parameters \"\$ERRSTACK_REF\" writers=syslog
    ade_error \"\$ERRSTACK_REF\" ADE_ERR_MISC \"$RANDOM_TEXT\"
    return \$ADE_FAIL
"
    #  We can't grep /var/log/syslog for $RANDOM_TEXT as file is not world-readable.
}

default_perl()
{
    adeperlf '
    ADE::error($errstack_ref, "ade_err_misc", "this is a test"); 
    return $ADE::FAIL;
'
}

devnull_perl()
{
    adeperlf '
    ADE::set_messaging_parameters($errstack_ref, writers=>["devnull"]);
    ADE::error($errstack_ref, "ade_err_misc", "this is a test"); 
    return $ADE::FAIL;
'
}

logfile_perl()
{
    LOGFILE=$(mktemp)
    adeperlf "
    ADE::set_messaging_parameters(\$errstack_ref, writers=>[\"logfile\"], logfile=>\"$LOGFILE\");
    ADE::error(\$errstack_ref, \"ade_err_misc\", \"this is a test\"); 
    return \$ADE::FAIL;
"
    #  Show file but trim timestamps.
    sed -r 's/^.{21}//' $LOGFILE
    rm $LOGFILE
}

syslog_perl()
{
    RANDOM_TEXT=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 20)
    adeperlf "
    ADE::set_messaging_parameters(\$errstack_ref, writers=>[\"syslog\"]);
    ADE::error(\$errstack_ref, \"ade_err_misc\", \"$RANDOM_TEXT\"); 
    return \$ADE::FAIL;
"
    #  We can't grep /var/log/syslog for $RANDOM_TEXT as file is not world-readable.
}

default_python()
{
    adepyf '
    import ade; ade.error(errstack, "ade_err_misc", "this is a test")
    return ade.fail
'
}

devnull_python()
{
    adepyf '
    import ade
    ade.set_messaging_parameters(errstack, stack=errstack, writers=["devnull"])
    ade.error(errstack, "ade_err_misc", "this is a test")
    return ade.fail
'
}

logfile_python()
{
    LOGFILE=$(mktemp)
    adepyf "
    import ade
    ade.set_messaging_parameters(errstack, stack=errstack, writers=[\"logfile\"], logfile=\"$LOGFILE\")
    ade.error(errstack, \"ade_err_misc\", \"this is a test\")
    return ade.fail
"
    #  Show file but trim timestamps.
    sed -r 's/^.{21}//' $LOGFILE
    rm $LOGFILE
}

syslog_python()
{
    RANDOM_TEXT=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 20)
    adepyf "
    import ade
    ade.set_messaging_parameters(errstack, stack=errstack, writers=[\"syslog\"])
    ade.error(errstack, \"ade_err_misc\", \"$RANDOM_TEXT\")
    return ade.fail
"
    #  We can't grep /var/log/syslog for $RANDOM_TEXT as file is not world-readable.
}

for CHECK in "${CHECKS[@]}"; do
    for ADELANG in "${ADELANGS[@]}"; do
        echo "$ADELANG/$CHECK:"
        ${CHECK}_$ADELANG
    done
done

echo "all done"
